"""Provide an asynchronous equivalent to the python console."""

import sys
import argparse
import shlex

from . import console


class AsynchronousCli(console.AsynchronousConsole):
    def __init__(
        self, commands, streams=None, *, prog=None, prompt_control=None, loop=None
    ):
        super().__init__(streams=streams, prompt_control=prompt_control, loop=loop)
        self.prog = prog
        self.commands = dict(commands)
        self.commands["help"] = (
            self.help_command,
            argparse.ArgumentParser(description="Display the help message."),
        )
        self.commands["list"] = (
            self.list_command,
            argparse.ArgumentParser(description="Display the command list."),
        )
        self.commands["exit"] = (
            self.exit_command,
            argparse.ArgumentParser(description="Exit the interface."),
        )
        for key, (corofunc, parser) in self.commands.items():
            parser.prog = key
            parser.print_help = lambda file=sys.stderr, *, self=parser: type(
                parser
            ).print_help(self, file)

    def get_default_banner(self):
        prog = self.prog or sys.argv[0].split("/")[-1]
        msg = f"Welcome to the CLI interface of {prog}!\n"
        msg += "Try:\n"
        msg += " * 'help' to display the help message\n"
        msg += " * 'list' to display the command list."
        return msg

    async def help_command(self, reader, writer):
        return """\
Type 'help' to display this message.
Type 'list' to display the command list.
Type '<command> -h' to display the help message of <command>."""

    async def list_command(self, reader, writer):
        msg = "List of commands:"
        for key, (corofunc, parser) in sorted(self.commands.items()):
            usage = parser.format_usage().replace("usage: ", "")[:-1]
            msg += "\n * " + usage
        return msg

    async def exit_command(self, reader, writer):
        raise SystemExit

    async def runsource(self, source, filename=None):
        # Parse the source
        if source.strip().endswith("\\"):
            return True
        source = source.replace("\\\n", "")
        try:
            name, *args = shlex.split(source)
        except ValueError:
            return False

        # Get the command
        if name not in self.commands:
            self.write(f"Command '{name}' does not exist.\n")
            await self.flush()
            return False
        corofunc, parser = self.commands[name]

        # Patch print_message so the parser prints to our console
        parser._print_message = lambda message, file=None: message and self.write(
            message
        )

        # Parse arguments
        try:
            namespace = parser.parse_args(args)
        except SystemExit:
            return False

        # Run the coroutine
        coro = corofunc(self.reader, self.writer, **vars(namespace))
        try:
            result = await coro
        except SystemExit:
            raise

        # Prompt the traceback or result
        except BaseException:
            self.showtraceback()
        else:
            if result is not None:
                self.write(str(result) + "\n")
        await self.flush()
        return False



"""Provide an asynchronous equivalent *to exec*."""

import ast
import codeop
from io import StringIO
from tokenize import generate_tokens, STRING, TokenError

CORO_NAME = "__corofn"
CORO_DEF = f"async def {CORO_NAME}(): "
CORO_CODE = CORO_DEF + "return (None, locals())\n"


def make_arg(key, annotation=None):
    """Make an ast function argument."""
    arg = ast.arg(key, annotation)
    arg.lineno, arg.col_offset = 0, 0
    return arg


def full_update(dct, values):
    """Fully update a dictionary."""
    dct.clear()
    dct.update(values)


def exec_single_result(obj, local, stream):
    """Reproduce the exec behavior in single mode (print and builtins._)"""
    local["_"] = obj
    if obj is not None:
        print(repr(obj), file=stream)


class ReturnChecker(ast.NodeVisitor):
    def __init__(self, filename):
        super().__init__()
        self.filename = filename

    def visit_FunctionDef(self, node: ast.FunctionDef):
        return  # skip functions

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        return  # async functions too

    def visit_Return(self, node: ast.Return):
        raise SyntaxError(
            "'return' outside function",
            (self.filename, node.lineno, node.col_offset + 1, None),
        )

    def visit_Yield(self, node: ast.Yield):
        raise SyntaxError(
            "'yield' outside function",
            (self.filename, node.lineno, node.col_offset + 1, None),
        )

    def visit_YieldFrom(self, node: ast.YieldFrom):
        self.visit_Yield(node)  # handle in the same way as regular yield


def make_tree(statement, filename, mode):
    """Helper for *aexec*."""
    # Check for returns and yields
    ReturnChecker(filename).visit(statement)

    # Create tree
    tree = ast.parse(CORO_CODE, filename, "single")
    # Check expression statement
    if isinstance(statement, ast.Expr):
        tree.body[0].body[0].value.elts[0] = statement.value
    else:
        tree.body[0].body.insert(0, statement)
    # Check the coroutine function
    exec(compile(tree, filename, "single"))

    if mode == "exec":
        return ast.Module([tree])
    if mode == "single":
        return ast.Interactive([tree])
    assert mode == "eval"
    raise ValueError("Mode 'eval' is not supported")


def make_coroutine_from_tree(wrapped, filename, local):
    """Make a coroutine from a tree structure."""
    dct = {}
    tree = wrapped.body[0]
    tree.body[0].args.args = list(map(make_arg, local))
    exec(compile(tree, filename, "single"), dct)
    return dct[CORO_NAME](**local)


def get_non_indented_lines(source):
    try:
        for token in generate_tokens(StringIO(source).readline):
            if token.type == STRING:
                # .start and .end line numbers are one-indexed
                yield from range(token.start[0], token.end[0])
    except TokenError:
        pass


def compile_for_aexec(
    source, filename, mode, dont_imply_dedent=False, local={}, **kwargs
):
    """Return a list of (coroutine object, abstract base tree)."""
    flags = ast.PyCF_ONLY_AST
    if dont_imply_dedent:
        flags |= codeop.PyCF_DONT_IMPLY_DEDENT
        # This flag is not available for python before 3.10
        try:
            flags |= codeop.PyCF_ALLOW_INCOMPLETE_INPUT
        except AttributeError:
            pass

    # Avoid a syntax error by wrapping code with `async def`
    # Disabling indentation inside multiline strings
    non_indented = set(  # sets are faster for `in` operation
        get_non_indented_lines(source)
    )
    indented = "\n".join(
        (" " * 4 if i not in non_indented and line else "") + line
        for i, line in enumerate(source.split("\n"))
    )
    coroutine = CORO_DEF + "\n" + indented + "\n"

    # Compilation is always performed in single mode
    compiled = compile(coroutine, filename, "single", flags)
    statements = compiled.body[0].body

    # Use original source to detect missing newlines, depending on the mode
    try:
        compile(source, filename, mode, flags)
    except SyntaxError:
        raise

    return [make_tree(statement, filename, mode) for statement in statements]


async def aexec(source, local=None, stream=None, filename="<aexec>"):
    """Asynchronous equivalent to *exec*."""
    if local is None:
        local = {}
    if isinstance(source, str):
        source = compile_for_aexec(source, filename, "exec")
    for tree in source:
        coro = make_coroutine_from_tree(tree, filename, local=local)
        result, new_local = await coro
        if isinstance(tree, ast.Interactive):
            exec_single_result(result, new_local, stream)
        full_update(local, new_local)


async def aeval(source, local=None):
    """Asynchronous equivalent to *eval*."""
    if local is None:
        local = {}

    if not isinstance(local, dict):
        raise TypeError("globals must be a dict")

    # Ensure that the result key is unique within the local namespace
    key = "__aeval_result__"
    while key in local:
        key += "_"

    # Perform syntax check to ensure the input is a valid eval expression
    try:
        ast.parse(source, mode="eval")
    except SyntaxError:
        raise

    # Assign the result of the expression to a known variable
    wrapped_code = f"{key} = {source}"

    # Use aexec to evaluate the wrapped code within the given local namespace
    await aexec(wrapped_code, local=local)

    # Return the result from the local namespace
    return local.pop(key)



import io
import asyncio

import pytest

from aioconsole import compat
from aioconsole.server import start_console_server, print_server


@pytest.mark.asyncio
async def test_server():
    server = await start_console_server(host="127.0.0.1", port=0, banner="test")
    address = server.sockets[0].getsockname()

    stream = io.StringIO()
    print_server(server, "test console", file=stream)
    expected = f"The test console is being served on 127.0.0.1:{address[1]}\n"
    assert stream.getvalue() == expected

    reader, writer = await asyncio.open_connection(*address)
    assert (await reader.readline()) == b"test\n"
    writer.write(b"1+1\n")
    assert (await reader.readline()) == b">>> 2\n"
    writer.write_eof()
    assert (await reader.readline()) == b">>> \n"
    writer.close()
    await writer.wait_closed()
    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_uds_server(tmpdir_factory):
    path = str(tmpdir_factory.mktemp("uds") / "my_uds")

    # Not available on windows
    if compat.platform == "win32":
        with pytest.raises(ValueError):
            await start_console_server(path=path, banner="test")
        return

    server = await start_console_server(path=path, banner="test")

    stream = io.StringIO()
    print_server(server, "test console", file=stream)
    expected = f"The test console is being served on {path}\n"
    assert stream.getvalue() == expected

    address = server.sockets[0].getsockname()
    reader, writer = await asyncio.open_unix_connection(address)
    assert (await reader.readline()) == b"test\n"
    writer.write(b"1+1\n")
    assert (await reader.readline()) == b">>> 2\n"
    writer.write_eof()
    assert (await reader.readline()) == b">>> \n"
    writer.close()
    await writer.wait_closed()
    server.close()
    await server.wait_closed()


@pytest.mark.asyncio
async def test_invalid_server():
    with pytest.raises(ValueError):
        await start_console_server()
    with pytest.raises(ValueError):
        await start_console_server(path="uds", port=0)



"""Provide the apython script."""

import os
import sys
import ast
import runpy
import warnings
import argparse
import traceback

from . import events
from . import server
from . import rlwrap
from . import compat

ZERO_WIDTH_SPACE = "\u200b"

DESCRIPTION = """\
Run the given python file or module with a modified asyncio policy replacing
the default event loop with an interactive loop.
If no argument is given, it simply runs an asynchronous python console."""

USAGE = """\
usage: apython [-h] [--serve [HOST:] PORT] [--no-readline]
               [--banner BANNER] [--locals LOCALS]
               [-m MODULE | FILE] ...
""".split(
    "usage: "
)[
    1
]


def exec_pythonstartup(locals_dict):
    filename = os.environ.get("PYTHONSTARTUP")
    if filename:
        if os.path.isfile(filename):
            with open(filename) as fobj:
                startupcode = fobj.read()
            try:
                locals_dict["__file__"] = filename
                exec(startupcode, globals(), locals_dict)
            except Exception:  # pragma: no cover
                traceback.print_exc()
            finally:
                locals_dict.pop("__file__", None)

        else:
            print(f"Could not open PYTHONSTARTUP - No such file: {filename}")


def parse_args(args=None):
    parser = argparse.ArgumentParser(
        prog="apython", description=DESCRIPTION, usage=USAGE
    )

    # Options

    parser.add_argument(
        "--serve",
        "-s",
        metavar="[HOST:] PORT",
        help="serve a console on the given interface instead",
    )
    parser.add_argument(
        "--no-readline",
        dest="readline",
        action="store_false",
        help="disable readline support",
    )
    parser.add_argument("--banner", help="provide a custom banner")
    parser.add_argument(
        "--locals", type=ast.literal_eval, help="provide custom locals as a dictionary"
    )

    # Hidden option

    parser.add_argument("--prompt-control", metavar="PC", help=argparse.SUPPRESS)

    # Input

    parser.add_argument("-m", dest="module", help="run a python module")
    parser.add_argument(
        "filename", metavar="FILE", nargs="?", help="python file to run"
    )

    # Extra arguments

    parser.add_argument(
        "args", metavar="ARGS", nargs=argparse.REMAINDER, help="extra arguments"
    )

    namespace = parser.parse_args(args)

    # If module is provided, filename is actually the fist arg
    if namespace.module is not None and namespace.filename is not None:
        namespace.args.insert(0, namespace.filename)

    # Parse the serve argument
    if namespace.serve is not None:
        namespace.serve = server.parse_server(namespace.serve, parser)

    return namespace


def load_readline():
    try:
        import readline  # noqa: F401
        import rlcompleter  # noqa: F401
    except ImportError:
        return False
    return True


def run_apython(args=None):
    namespace = parse_args(args)

    if (
        namespace.readline
        and not namespace.serve
        and compat.platform != "win32"
        and load_readline()
    ):
        # Run python interactive hook in order to configure binding and history support
        interactive_hook = getattr(sys, "__interactivehook__", None)
        if interactive_hook:
            try:
                interactive_hook()
            except Exception as exc:
                warnings.warn(f"Interactive hook failed: {exc!r}", stacklevel=2)

        code = run_apython_in_subprocess(args, namespace.prompt_control)
        sys.exit(code)

    try:
        sys._argv = sys.argv
        sys._path = sys.path
        if namespace.module:
            sys.argv = [None] + namespace.args
            sys.path.insert(0, "")
            events.set_interactive_policy(
                locals=namespace.locals,
                banner=namespace.banner,
                serve=namespace.serve,
                prompt_control=namespace.prompt_control,
            )
            runpy.run_module(namespace.module, run_name="__main__", alter_sys=True)
        elif namespace.filename:
            sys.argv = [None] + namespace.args
            path = os.path.dirname(os.path.abspath(namespace.filename))
            sys.path.insert(0, path)
            events.set_interactive_policy(
                locals=namespace.locals,
                banner=namespace.banner,
                serve=namespace.serve,
                prompt_control=namespace.prompt_control,
            )
            runpy.run_path(namespace.filename, run_name="__main__")
        else:
            if namespace.locals is None:
                namespace.locals = {}
            exec_pythonstartup(namespace.locals)
            events.run_console(
                locals=namespace.locals,
                banner=namespace.banner,
                serve=namespace.serve,
                prompt_control=namespace.prompt_control,
            )
    finally:
        sys.argv = sys._argv
        sys.path = sys._path

    sys.exit()


def run_apython_in_subprocess(args=None, prompt_control=None):
    # Default arguments
    if args is None:
        args = sys.argv[1:]
    if prompt_control is None:
        prompt_control = ZERO_WIDTH_SPACE
    # Create subprocess
    proc_args = [
        sys.executable,
        "-m",
        "aioconsole",
        "--no-readline",
        "--prompt-control",
        prompt_control,
    ]
    return rlwrap.rlwrap_process(proc_args + args, prompt_control, use_stderr=True)



"""Provide an asynchronous equivalent to *input*."""

import os
import sys
import stat
import weakref
import asyncio
import selectors
from collections import deque
from threading import Thread
from concurrent.futures import Future

from . import compat


class ProtectedPipe:
    """Wrapper to protect a pipe from being closed."""

    def __init__(self, pipe):
        self.pipe = pipe

    def fileno(self):
        return self.pipe.fileno()

    def close(self):
        pass


def is_pipe_transport_compatible(pipe):
    if compat.platform == "win32":
        return False
    try:
        fileno = pipe.fileno()
    except (OSError, AttributeError):
        return False
    mode = os.fstat(fileno).st_mode
    is_char = stat.S_ISCHR(mode)
    is_fifo = stat.S_ISFIFO(mode)
    is_socket = stat.S_ISSOCK(mode)
    if not (is_char or is_fifo or is_socket):
        return False
    # Fail early when the file descriptor cannot be registered.
    # This happens with docker containers for instance.
    # See issue #102: https://github.com/vxgmichel/aioconsole/issues/102
    try:
        with selectors.DefaultSelector() as selector:
            selector.register(fileno, selectors.EVENT_READ | selectors.EVENT_WRITE)
    except OSError:
        return False
    return True


async def run_as_daemon(func, *args):
    future = Future()
    future.set_running_or_notify_cancel()

    # A bug in python 3.7 makes it a bad idea to set a BaseException
    # in a wrapped future (see except statement in asyncio.Task.__wakeup)
    # Instead, we'll wrap base exceptions into exceptions and unwrap them
    # on the other side of the call.
    class BaseExceptionWrapper(Exception):
        pass

    def daemon():
        try:
            result = func(*args)
        except Exception as e:
            future.set_exception(e)
        except BaseException as e:
            future.set_exception(BaseExceptionWrapper(e))
        else:
            future.set_result(result)

    Thread(target=daemon, daemon=True).start()
    try:
        return await asyncio.wrap_future(future)
    except BaseExceptionWrapper as exc:
        raise exc.args[0]


class StandardStreamReaderProtocol(asyncio.StreamReaderProtocol):
    def connection_made(self, transport):
        # The connection is already made
        if self._stream_reader._transport is not None:
            return
        # Make the connection
        super().connection_made(transport)

    def connection_lost(self, exc):
        # Copy the inner state
        state = self.__dict__.copy()
        # Call the parent
        super().connection_lost(exc)
        # Restore the inner state
        self.__dict__.update(state)


class StandardStreamReader(asyncio.StreamReader):
    async def readuntil(self, separator=b"\n"):
        # Re-implement `readuntil` to work around self._limit.
        # The limit is still useful to prevent the internal buffer
        # from growing too large when it's not necessary, but it
        # needs to be disabled when the user code is purposely
        # reading from stdin.
        while True:
            try:
                return await super().readuntil(separator)
            except asyncio.LimitOverrunError as e:
                if self._buffer.startswith(separator, e.consumed):
                    chunk = self._buffer[: e.consumed + len(separator)]
                    del self._buffer[: e.consumed + len(separator)]
                    self._maybe_resume_transport()
                    return bytes(chunk)
                await self._wait_for_data("readuntil")


class StandardStreamWriter(asyncio.StreamWriter):
    def __del__(self):
        # No `__del__` method for StreamWriter in Python 3.10 and before
        try:
            parent_del = super().__del__
        except AttributeError:
            return
        # Do not attempt to close the transport if the loop is closed
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return
        parent_del()

    def write(self, data):
        if isinstance(data, str):
            data = data.encode()
        super().write(data)


class NonFileStreamReader:
    def __init__(self, stream, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.stream = stream
        self.eof = False

    def at_eof(self):
        return self.eof

    async def readline(self):
        try:
            data = await run_as_daemon(self.stream.readline)
        except AttributeError:
            raise RuntimeError("ainput(): lost sys.stdin")
        if isinstance(data, str):
            data = data.encode()
        self.eof = not data
        return data

    async def read(self, n=-1):
        try:
            data = await run_as_daemon(self.stream.read, n)
        except AttributeError:
            raise RuntimeError("ainput(): lost sys.stdin")
        if isinstance(data, str):
            data = data.encode()
        self.eof = not data
        return data

    def __aiter__(self):
        return self

    async def __anext__(self):
        val = await self.readline()
        if val == b"":
            raise StopAsyncIteration
        return val


class NonFileStreamWriter:
    def __init__(self, stream, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self.stream = stream
        self.buffer = deque()
        self.write_task = None
        self.task_finalizer = None

    def write(self, data):
        if self.stream is None:
            raise RuntimeError("This writer stream is already closed")
        if isinstance(data, bytes):
            data = data.decode()
        self.buffer.append(data)
        if self.write_task is not None and not self.write_task.done():
            return
        if self.write_task is not None and self.write_task.done():
            self.write_task = None
            self.task_finalizer()
        self.write_task = asyncio.ensure_future(
            _nonfile_stream_writer_task_target(self.buffer, self.stream)
        )
        self.task_finalizer = weakref.finalize(self, self.write_task.result)

    async def drain(self):
        if self.write_task is not None:
            try:
                await self.write_task
            finally:
                self.write_task = None
                self.task_finalizer.detach()

    def close(self):
        self.stream = None

    def is_closing(self):
        return self.stream is None and self.write_task is not None

    async def wait_closed(self):
        await self.drain()


async def _nonfile_stream_writer_task_target(data_buffer, stream):
    loop = asyncio.get_event_loop()
    while data_buffer:
        data = data_buffer.popleft()
        await loop.run_in_executor(None, stream.write, data)
    if hasattr(stream, "flush"):
        await loop.run_in_executor(None, stream.flush)


async def open_standard_pipe_connection(pipe_in, pipe_out, pipe_err, *, loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()

    # Reader
    in_reader = StandardStreamReader(loop=loop)
    protocol = StandardStreamReaderProtocol(in_reader, loop=loop)
    await loop.connect_read_pipe(lambda: protocol, ProtectedPipe(pipe_in))

    # Out writer
    out_transport, _ = await loop.connect_write_pipe(
        lambda: protocol, ProtectedPipe(pipe_out)
    )
    out_writer = StandardStreamWriter(out_transport, protocol, in_reader, loop)

    # Err writer
    err_transport, _ = await loop.connect_write_pipe(
        lambda: protocol, ProtectedPipe(pipe_err)
    )
    err_writer = StandardStreamWriter(err_transport, protocol, in_reader, loop)

    # Set the write buffer limits to zero
    # This way, `await stream.drain()` can be used to make sure the buffer is flushed
    out_transport.set_write_buffer_limits(high=0, low=0)
    err_transport.set_write_buffer_limits(high=0, low=0)

    # Return
    return in_reader, out_writer, err_writer


async def create_standard_streams(stdin, stdout, stderr, *, loop=None):
    if all(map(is_pipe_transport_compatible, (stdin, stdout, stderr))):
        return await open_standard_pipe_connection(stdin, stdout, stderr, loop=loop)
    return (
        NonFileStreamReader(stdin, loop=loop),
        NonFileStreamWriter(stdout, loop=loop),
        NonFileStreamWriter(stderr, loop=loop),
    )


async def get_standard_streams(*, cache={}, use_stderr=False, loop=None):
    if loop is None:
        loop = asyncio.get_event_loop()
    args = sys.stdin, sys.stdout, sys.stderr
    key = args, loop
    if cache.get(key) is None:
        cache[key] = await create_standard_streams(*args, loop=loop)
    in_reader, out_writer, err_writer = cache[key]
    return in_reader, err_writer if use_stderr else out_writer


async def ainput(prompt="", *, streams=None, use_stderr=False, loop=None):
    """Asynchronous equivalent to *input*."""
    # Get standard streams
    if streams is None:
        streams = await get_standard_streams(use_stderr=use_stderr, loop=loop)
    reader, writer = streams
    # Write prompt
    writer.write(prompt.encode())
    await writer.drain()
    # Get data
    data = await reader.readline()
    # Decode data
    data = data.decode()
    # Return or raise EOF
    if not data.endswith("\n"):
        raise EOFError
    return data.rstrip("\n")


async def aprint(
    *values, sep=None, end="\n", flush=True, streams=None, use_stderr=False, loop=None
):
    """Asynchronous equivalent to *print*."""
    # Get standard streams
    if streams is None:
        streams = await get_standard_streams(use_stderr=use_stderr, loop=loop)
    _, writer = streams

    print(*values, sep=sep, end=end, flush=False, file=writer)

    if flush:
        await writer.drain()



import io
import os
import platform
import sys
import signal
import asyncio
from unittest.mock import Mock
from contextlib import contextmanager

import pytest
from aioconsole import interact
from aioconsole.stream import NonFileStreamReader, NonFileStreamWriter


@contextmanager
def stdcontrol(monkeypatch):
    # PS1
    monkeypatch.setattr("sys.ps1", "[Hello!]", raising=False)
    # Stdin control
    stdin_read, stdin_write = os.pipe()
    monkeypatch.setattr("sys.stdin", open(stdin_read))
    writer = NonFileStreamWriter(open(stdin_write, "w"))
    # Stdout control
    monkeypatch.setattr(sys, "stdout", io.StringIO())
    # Stderr control
    stderr_read, stderr_write = os.pipe()
    monkeypatch.setattr("sys.stderr", open(stderr_write, "w"))
    reader = NonFileStreamReader(open(stderr_read))
    # Yield
    yield reader, writer
    # Check
    assert sys.stdout.getvalue() == ""


async def assert_stream(stream, expected, loose=False):
    s = None if loose else "\n"
    for expected_line in expected.splitlines():
        line = await stream.readline()
        assert expected_line.strip(s) == line.decode().strip(s)


@pytest.fixture(params=["unix", "not-unix"])
async def signaling(request, monkeypatch):
    if request.param == "not-unix":
        event_loop = asyncio.get_running_loop()
        m = Mock(side_effect=NotImplementedError)
        monkeypatch.setattr(event_loop, "add_signal_handler", m)
        monkeypatch.setattr(event_loop, "remove_signal_handler", m)
    yield request.param


@pytest.mark.asyncio
async def test_interact_simple(monkeypatch):
    with stdcontrol(monkeypatch) as (reader, writer):
        banner = "A BANNER"
        writer.write("1+1\n")
        await writer.drain()
        writer.stream.close()
        await interact(banner=banner, stop=False)
        await assert_stream(reader, banner)
        await assert_stream(reader, sys.ps1 + "2")
        await assert_stream(reader, sys.ps1)


@pytest.mark.asyncio
async def test_interact_traceback(monkeypatch):
    with stdcontrol(monkeypatch) as (reader, writer):
        banner = "A BANNER"
        writer.write("1/0\n")
        await writer.drain()
        writer.stream.close()
        await interact(banner=banner, stop=False)
        # Check stderr
        await assert_stream(reader, banner)
        await assert_stream(reader, sys.ps1 + "Traceback (most recent call last):")
        # Skip 3 lines
        for _ in range(4 if sys.version_info >= (3, 11) else 3):
            await reader.readline()
        # Check stderr
        await assert_stream(reader, "ZeroDivisionError: division by zero")
        await assert_stream(reader, sys.ps1)


@pytest.mark.asyncio
async def test_interact_syntax_error(monkeypatch):
    with stdcontrol(monkeypatch) as (reader, writer):
        writer.write("a b\n")
        await writer.drain()
        writer.stream.close()
        banner = "A BANNER"
        await interact(banner=banner, stop=False)
        await assert_stream(reader, banner)
        # Skip line
        await reader.readline()
        await assert_stream(reader, "    a b")
        if sys.version_info >= (3, 10, 0) and sys.version_info < (3, 10, 1):
            await assert_stream(reader, "    ^^^", loose=True)
        else:
            await assert_stream(reader, "      ^", loose=True)
        if (
            sys.version_info >= (3, 10, 0)
            and sys.version_info < (3, 10, 1)
            or (
                platform.python_implementation() == "PyPy"
                and sys.version_info >= (3, 9)
                and sys.version_info < (3, 10)
            )
        ):
            await assert_stream(
                reader, "SyntaxError: invalid syntax. Perhaps you forgot a comma?"
            )
        else:
            await assert_stream(reader, "SyntaxError: invalid syntax")
        await assert_stream(reader, sys.ps1)


@pytest.mark.asyncio
async def test_interact_keyboard_interrupt(monkeypatch, signaling):
    with stdcontrol(monkeypatch) as (reader, writer):
        # Start interaction
        banner = "A BANNER"
        task = asyncio.ensure_future(interact(banner=banner, stop=False))
        # Wait for banner
        await assert_stream(reader, banner)
        # Send SIGINT
        if sys.platform == "win32":
            signal.getsignal(signal.SIGINT)(signal.SIGINT, None)
        else:
            os.kill(os.getpid(), signal.SIGINT)
        # Wait for ps1
        await assert_stream(reader, sys.ps1)
        await assert_stream(reader, "KeyboardInterrupt")
        # Close stdin
        writer.stream.close()
        # Wait for interact to finish
        await assert_stream(reader, sys.ps1)
        await task
        # Test
        assert sys.stdout.getvalue() == ""


@pytest.mark.asyncio
async def test_broken_pipe(monkeypatch, signaling):
    with stdcontrol(monkeypatch) as (reader, writer):
        # Start interaction
        banner = "A BANNER"
        task = asyncio.ensure_future(interact(banner=banner, stop=False))
        # Wait for banner
        await assert_stream(reader, banner)
        # Close stdin
        writer.stream.close()
        reader.stream.close()
        # Wait for interact to finish
        await task
        # Test
        assert sys.stdout.getvalue() == ""


@pytest.mark.asyncio
async def test_interact_multiple_indented_lines(monkeypatch):
    with stdcontrol(monkeypatch) as (reader, writer):
        banner = "A BANNER"
        writer.write("def x():\n    print(1)\n    print(2)\n\nx()\n")
        await writer.drain()
        writer.stream.close()
        await interact(banner=banner, stop=False)
        await assert_stream(reader, banner)
        await assert_stream(reader, sys.ps1 + sys.ps2 * 3 + sys.ps1 + "1\n2")


@pytest.mark.asyncio
async def test_interact_cancellation(monkeypatch):
    with stdcontrol(monkeypatch) as (reader, writer):
        banner = "A BANNER"
        task = asyncio.ensure_future(interact(banner=banner, stop=False))
        # Wait for banner
        await assert_stream(reader, banner)
        task.cancel()
        with pytest.raises(asyncio.CancelledError):
            await task
        assert task.cancelled



import os
import io
import gc
import sys
import pytest
import asyncio
from unittest.mock import Mock

from aioconsole.stream import create_standard_streams, ainput, aprint
from aioconsole.stream import is_pipe_transport_compatible, get_standard_streams


@pytest.mark.skipif(sys.platform == "win32", reason="Not supported on windows")
@pytest.mark.asyncio
async def test_create_standard_stream_with_pipe(is_uvloop):
    r1, w1 = os.pipe()
    r2, w2 = os.pipe()
    stdin = open(r1)
    stdout = open(w1, "w")
    stderr = open(w2, "w")

    assert is_pipe_transport_compatible(stdin)
    assert is_pipe_transport_compatible(stdout)
    assert is_pipe_transport_compatible(stderr)

    reader, writer1, writer2 = await create_standard_streams(stdin, stdout, stderr)

    writer1.write("a\n")
    await writer1.drain()
    data = await reader.readline()
    assert data == b"a\n"

    writer2.write("b\n")
    await writer2.drain()
    assert os.read(r2, 2) == b"b\n"

    # Mock stdout.close() to check if it has been called
    stdout_actual_close = stdout.close
    stdout.close = Mock()

    # Close the transport and delete the object
    writer1.transport.close()
    del reader, writer1, writer2
    gc.collect()  # Force garbage collection - necessary for pypy

    # Check that the transport has been closed but not stdout
    assert not stdout.close.called

    # Weirdly enough, uvloop DID close the file descriptor.
    # Probably because it used the file descriptor directly instead
    # of the pipe object. However, this should not be an issue since
    # file descriptors 0, 1, 2 do not seem to be affected by this.
    if not is_uvloop:
        stdout_actual_close()


@pytest.mark.asyncio
async def test_create_standard_stream_with_non_pipe(monkeypatch):
    stdin = io.StringIO("a\nb\nc\nd\n")
    stdout = io.StringIO()
    stderr = io.StringIO()
    reader, writer1, writer2 = await create_standard_streams(stdin, stdout, stderr)

    writer1.write("a\n")
    await writer1.drain()
    data = await reader.readline()
    assert data == b"a\n"
    assert stdout.getvalue() == "a\n"

    writer2.write("b\n")
    await writer2.drain()
    data = await reader.readline()
    assert data == b"b\n"
    assert stderr.getvalue() == "b\n"

    # Multiple writes
    writer2.write("c\n")
    writer2.write("d\n")
    await asyncio.sleep(0.1)
    writer2.write("e\n")
    writer2.close()
    assert writer2.is_closing()
    await writer2.wait_closed()
    assert stderr.getvalue() == "b\nc\nd\ne\n"
    with pytest.raises(RuntimeError):
        writer2.write("f\n")

    data = await reader.read(2)
    assert data == b"c\n"

    assert reader.at_eof() is False

    async for data in reader:
        assert data == b"d\n"

    assert reader.at_eof() is True

    # Check exception handling in the daemon thread

    class KeyboardInterruptLike(BaseException):
        pass

    def raise_keyboard_interrupt(*args):
        raise KeyboardInterruptLike

    def raise_os_error(*args):
        raise OSError

    monkeypatch.setattr(stdin, "readline", raise_os_error)
    with pytest.raises(OSError):
        data = await reader.readline()

    monkeypatch.setattr(stdin, "read", raise_keyboard_interrupt)
    with pytest.raises(KeyboardInterruptLike):
        data = await reader.read()


def mock_stdio(monkeypatch, input_text="", disable_stdin=False):
    monkeypatch.setattr("sys.stdin", None if disable_stdin else io.StringIO(input_text))
    monkeypatch.setattr("sys.stdout", io.StringIO())
    monkeypatch.setattr("sys.stderr", io.StringIO())


@pytest.mark.asyncio
async def test_ainput_with_standard_stream(monkeypatch):
    mock_stdio(monkeypatch, "a\nb\n")
    assert (await ainput()) == "a"
    assert (await ainput(">>> ")) == "b"
    assert sys.stdout.getvalue() == ">>> "
    assert sys.stderr.getvalue() == ""


@pytest.mark.asyncio
async def test_aprint_with_standard_stream(monkeypatch):
    mock_stdio(monkeypatch)
    await aprint("ab", "cd")
    assert sys.stdout.getvalue() == "ab cd\n"
    await aprint("a" * 1024 * 64)
    assert sys.stdout.getvalue() == "ab cd\n" + "a" * 1024 * 64 + "\n"
    assert sys.stderr.getvalue() == ""


@pytest.mark.parametrize("flush", [False, True])
@pytest.mark.asyncio
async def test_aprint_flush_argument(monkeypatch, flush):
    mock_stdio(monkeypatch)
    await aprint("a", flush=flush)
    if not flush:
        # Might or might not be there yet, depending on internal logic
        assert sys.stdout.getvalue() in ("", "a\n")
        await aprint("", end="", flush=True)
    assert sys.stdout.getvalue() == "a\n"


@pytest.mark.asyncio
async def test_read_from_closed_pipe(is_uvloop):
    if is_uvloop:
        pytest.skip("This test is flaky with uvloop for some reason.")

    stdin_r, stdin_w = os.pipe()
    stdout_r, stdout_w = os.pipe()
    stderr_r, stderr_w = os.pipe()

    stdin = open(stdin_w, "wb")
    stdin.write(b"hello\n")
    stdin.close()

    f_stdin = open(stdin_r, "r")
    f_stdout = open(stdout_w, "w")
    f_stderr = open(stderr_w, "w")

    reader, writer1, writer2 = await create_standard_streams(
        f_stdin, f_stdout, f_stderr
    )

    result = await ainput(">>> ", streams=(reader, writer1))
    assert result == "hello"

    writer1.close()
    await writer1.wait_closed()
    f_stdout.close()

    writer2.close()
    await writer2.wait_closed()
    f_stderr.close()

    assert open(stdout_r).read() == ">>> "
    assert open(stderr_r).read() == ""


@pytest.mark.skipif(sys.platform == "win32", reason="Not supported on windows")
@pytest.mark.asyncio
async def test_standard_stream_pipe_buffering(is_uvloop):
    if is_uvloop:
        pytest.skip("This test is flaky with uvloop for some reason.")
    r1, w1 = os.pipe()
    r2, w2 = os.pipe()
    stdin = open(r1)
    stdout = open(w1, "w")
    stderr = open(w2, "w")

    assert is_pipe_transport_compatible(stdin)
    assert is_pipe_transport_compatible(stdout)
    assert is_pipe_transport_compatible(stderr)

    reader, writer1, writer2 = await create_standard_streams(stdin, stdout, stderr)

    blob_size = 4 * 1024 * 1024  # 4 MB
    writer1.write("a\n" + "b" * blob_size + "\n")
    task = asyncio.ensure_future(writer1.drain())
    data = await reader.readline()
    assert data == b"a\n"

    # Check back pressure
    await asyncio.sleep(0.1)
    assert not task.done()
    assert len(reader._buffer) < blob_size

    data = await reader.readline()
    assert data == b"b" * blob_size + b"\n"
    await task


@pytest.mark.asyncio
async def test_aprint_with_no_stdin(monkeypatch):
    mock_stdio(monkeypatch, disable_stdin=True)
    await aprint("test1")
    assert sys.stdout.getvalue() == "test1\n"
    assert sys.stderr.getvalue() == ""
    with pytest.raises(RuntimeError) as ctx:
        await ainput("test2")
    assert str(ctx.value) == "ainput(): lost sys.stdin"
    assert sys.stdout.getvalue() == "test1\ntest2"
    assert sys.stderr.getvalue() == ""
    # Test the methods specifically
    reader, _ = await get_standard_streams()
    with pytest.raises(RuntimeError) as ctx:
        await reader.read(10)
    assert str(ctx.value) == "ainput(): lost sys.stdin"
    with pytest.raises(RuntimeError) as ctx:
        await reader.readline()
    assert str(ctx.value) == "ainput(): lost sys.stdin"



"""Provide a readline wrapper to control a subprocess."""

import sys
import ctypes
import signal
import builtins
import subprocess
from concurrent.futures import ThreadPoolExecutor

from . import compat

if compat.platform == "darwin":
    import fcntl


def rlwrap_process(args, prompt_control, use_stderr=False):
    assert len(prompt_control) == 1
    # Start process
    process = subprocess.Popen(
        args,
        bufsize=0,
        universal_newlines=True,
        stdin=subprocess.PIPE,
        **{"stderr" if use_stderr else "stdout": subprocess.PIPE},
    )
    # Readline wrapping
    return _rlwrap(process, prompt_control, use_stderr)


def _rlwrap(process, prompt_control, use_stderr=False):
    # Get source and destination
    source = process.stderr if use_stderr else process.stdout
    dest = sys.stderr if use_stderr else sys.stdout

    # Check prompt control
    assert len(prompt_control) == 1

    # Run background task
    with ThreadPoolExecutor(1) as executor:
        future = executor.submit(wait_for_prompt, source, dest, prompt_control)

        # Loop over prompts
        while process.poll() is None:
            # Get prompt
            try:
                prompt = future.result()
            except KeyboardInterrupt:
                process.send_signal(signal.SIGINT)
                continue
            except EOFError:
                break
            else:
                future = executor.submit(wait_for_prompt, source, dest, prompt_control)

            # Get user input
            try:
                raw = input(prompt, use_stderr=use_stderr) + "\n"
            except KeyboardInterrupt:
                process.send_signal(signal.SIGINT)
                continue
            except EOFError:
                break
            else:
                process.stdin.write(raw)

        # Close and wait process streams
        process.stdin.close()
        future.exception()

    # Wait process and return code
    return process.wait()


def wait_for_prompt(src, dest, prompt_control, buffersize=1):
    def read():
        value = src.read(buffersize)
        if value:
            return value
        raise EOFError

    def write(arg):
        if arg:
            dest.write(arg)
            dest.flush()

    # Prevent exception in macOS with large output (issue #42)
    if compat.platform == "darwin":
        fcntl.fcntl(dest.fileno(), fcntl.F_SETFL, 0)

    # Wait for first prompt control
    while True:
        current = read()
        if prompt_control in current:
            break
        write(current)

    preprompt, current = current.split(prompt_control, 1)
    write(preprompt)

    # Wait for second prompt control
    while prompt_control not in current:
        current += read()

    prompt, postprompt = current.split(prompt_control, 1)
    write(postprompt)

    return prompt


def input(prompt="", use_stderr=False):
    # Use readline if possible
    try:
        import readline  # noqa
    except ImportError:
        return builtins.input(prompt)
    # Use stdout
    if not use_stderr:
        return builtins.input(prompt)
    api = ctypes.pythonapi
    # Cross-platform compatibility
    if compat.platform == "darwin":
        stdin = "__stdinp"
        stderr = "__stderrp"
    else:
        stdin = "stdin"
        stderr = "stderr"
    # Get standard streams
    try:
        fin = ctypes.c_void_p.in_dll(api, stdin)
        ferr = ctypes.c_void_p.in_dll(api, stderr)
    # Cygwin fallback
    except ValueError:
        return builtins.input(prompt)
    # Call readline
    call_readline = api.PyOS_Readline
    call_readline.restype = ctypes.c_char_p
    result = call_readline(fin, ferr, prompt.encode())
    # Decode result
    if len(result) == 0:
        raise EOFError
    return result.decode().rstrip("\n")



import io
import sys
import argparse

import pytest

from aioconsole import AsynchronousCli

testdata = {
    "simple_command": (
        "hello\n",
        """\
Welcome to the CLI interface of hello!
Try:
 * 'help' to display the help message
 * 'list' to display the command list.
[Hello!] Hello!
[Hello!] \n""",
    ),
    "simple_command_with_arg": (
        "hello -n Neil\n",
        """\
Welcome to the CLI interface of hello!
Try:
 * 'help' to display the help message
 * 'list' to display the command list.
[Hello!] Hello Neil!
[Hello!] \n""",
    ),
    "list_command": (
        "list\n",
        """\
Welcome to the CLI interface of hello!
Try:
 * 'help' to display the help message
 * 'list' to display the command list.
[Hello!] List of commands:
 * exit [-h]
 * hello [-h] [--name NAME]
 * help [-h]
 * list [-h]
[Hello!] \n""",
    ),
    "help_command": (
        "help\n",
        """\
Welcome to the CLI interface of hello!
Try:
 * 'help' to display the help message
 * 'list' to display the command list.
[Hello!] Type 'help' to display this message.
Type 'list' to display the command list.
Type '<command> -h' to display the help message of <command>.
[Hello!] \n""",
    ),
    "exit_command": (
        "exit\n",
        """\
Welcome to the CLI interface of hello!
Try:
 * 'help' to display the help message
 * 'list' to display the command list.
[Hello!] """,
    ),
    "wrong_command": (
        "hellooo\n",
        """\
Welcome to the CLI interface of hello!
Try:
 * 'help' to display the help message
 * 'list' to display the command list.
[Hello!] Command 'hellooo' does not exist.
[Hello!] \n""",
    ),
}


def make_cli(streams=None):
    async def say_hello(reader, writer, name=None):
        data = f"Hello {name}!" if name else "Hello!"
        writer.write(data.encode() + b"\n")

    parser = argparse.ArgumentParser(description="Say hello")
    parser.add_argument("--name", "-n", type=str)
    commands = {"hello": (say_hello, parser)}
    return AsynchronousCli(commands, streams, prog="hello")


@pytest.mark.parametrize(
    "input_string, expected", list(testdata.values()), ids=list(testdata.keys())
)
@pytest.mark.asyncio
async def test_async_cli(monkeypatch, input_string, expected):
    monkeypatch.setattr("sys.ps1", "[Hello!] ", raising=False)
    monkeypatch.setattr("sys.stdin", io.StringIO(input_string))
    monkeypatch.setattr("sys.stderr", io.StringIO())
    await make_cli().interact(stop=False)
    assert sys.stderr.getvalue() == expected



"""Serve the python console using socket communication."""

import asyncio
import socket
from functools import partial

from . import compat
from . import console


async def handle_connect(reader, writer, factory, banner=None):
    streams = reader, writer
    interface = factory(streams=streams)
    await interface.interact(banner=banner, stop=False, handle_sigint=False)
    writer.close()


async def start_interactive_server(
    factory=console.AsynchronousConsole,
    host=None,
    port=None,
    path=None,
    banner=None,
    *,
    loop=None,
):
    if compat.platform == "win32" and port is None:
        raise ValueError("A TCP port should be provided")
    if (port is None) == (path is None):
        raise ValueError("Either a TCP port or a UDS path should be provided")
    if port is not None:
        # Override asyncio behavior (i.e serve on all interfaces by default)
        host = host or "localhost"
        start_server = partial(asyncio.start_server, host=host, port=port)
    else:
        start_server = partial(asyncio.start_unix_server, path=path)

    client_connected = partial(handle_connect, factory=factory, banner=banner)
    server = await start_server(client_connected)
    return server


async def start_console_server(
    host=None,
    port=None,
    path=None,
    locals=None,
    filename="<console>",
    banner=None,
    prompt_control=None,
    *,
    loop=None,
):
    def factory(streams):
        client_locals = dict(locals) if locals is not None else None
        return console.AsynchronousConsole(
            streams=streams,
            locals=client_locals,
            filename=filename,
            prompt_control=prompt_control,
        )

    server = await start_interactive_server(
        factory, host=host, port=port, path=path, banner=banner, loop=loop
    )
    return server


def print_server(server, name="console", file=None):
    interface = server.sockets[0].getsockname()
    AF_UNIX = None if compat.platform == "win32" else socket.AF_UNIX
    if server.sockets[0].family != AF_UNIX:
        interface = "{}:{}".format(*interface)
    print(f"The {name} is being served on {interface}", file=file)


def run(host=None, port=None, path=None):
    loop = asyncio.get_event_loop()
    coro = start_interactive_server(host=host, port=port, path=path)
    loop.server = loop.run_until_complete(coro)
    print_server(loop.server)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass


def parse_server(server, parser=None):
    try:
        host, port = server.rsplit(":", maxsplit=1)
    except ValueError:
        host, port = "localhost", server
    try:
        port = int(port)
    except (ValueError, TypeError):
        msg = f"{server!r} is not a valid server [HOST:]PORT"
        if not parser:
            raise ValueError(msg)
        parser.error(msg)
    return host, port



import io
import sys
from contextlib import contextmanager

from unittest.mock import Mock, patch, call

import pytest

from aioconsole import InteractiveEventLoop, apython, rlwrap, compat


startupfile = """
def hehe():
    return 42

foo = 1

# Imports work
import math
r = math.cos(0)

# Exec works and is visible from the interpreter
s = 'from pprint import pprint'
exec(s)

"""


@contextmanager
def mock_module(name):
    try:
        module = sys.modules.get(name)
        sys.modules[name] = Mock()
        yield sys.modules[name]
    finally:
        if module is None:
            del sys.modules[name]
        else:
            sys.modules[name] = module


@pytest.fixture(params=["linux", "darwin", "win32"])
def platform(request):
    with patch("aioconsole.compat.platform", new=request.param):
        yield request.param


@pytest.fixture
def mock_readline(platform):
    with mock_module("readline"):
        with patch("aioconsole.rlwrap.ctypes") as m_ctypes:
            if platform == "darwin":
                stdin = "__stdinp"
                stderr = "__stderrp"
            else:
                stdin = "stdin"
                stderr = "stderr"

            def readline(fin, ferr, prompt):
                sys.stderr.write(prompt.decode())
                return sys.stdin.readline().encode()

            api = m_ctypes.pythonapi
            call_readline = api.PyOS_Readline
            call_readline.side_effect = readline

            if platform == "darwin":
                with patch("aioconsole.rlwrap.fcntl", create=True):
                    yield call_readline
            else:
                yield call_readline

            if call_readline.called:
                m_ctypes.c_void_p.in_dll.assert_has_calls(
                    [call(api, stdin), call(api, stderr)]
                )


@pytest.fixture(params=["readline", "no-readline"])
def use_readline(request, mock_readline, platform):
    if request.param == "readline":
        # Readline tests hang on windows for some reason
        if sys.platform == "win32":
            pytest.xfail()
        return []
    return ["--no-readline"]


def test_input_with_stderr_prompt(mock_readline):
    with patch("sys.stdin", new=io.StringIO("test\n")):
        assert rlwrap.input(use_stderr=True) == "test"


def test_basic_apython_usage(capfd, use_readline):
    with patch("sys.stdin", new=io.StringIO("1+1\n")):
        with pytest.raises(SystemExit):
            apython.run_apython(["--banner=test"] + use_readline)
    out, err = capfd.readouterr()
    assert out == ""
    assert err == "test\n>>> 2\n>>> \n"


def test_missing_readline(capfd, use_readline):
    sys.modules["readline"] = None
    with patch("sys.stdin", new=io.StringIO("1+1\n")):
        with pytest.raises(SystemExit):
            apython.run_apython(["--banner=test"] + use_readline)
    out, err = capfd.readouterr()
    assert out == ""
    assert err == "test\n>>> 2\n>>> \n"


def test_interactive_hook_error(capfd, use_readline, platform):
    sys.modules["readline"].parse_and_bind.side_effect = ZeroDivisionError()
    with patch("sys.stdin", new=io.StringIO("1+1\n")):
        with pytest.raises(SystemExit):
            with pytest.warns(UserWarning):
                apython.run_apython(["--banner=test"] + use_readline)
    out, err = capfd.readouterr()
    assert out == ""
    assert err == "test\n>>> 2\n>>> \n"


def test_basic_apython_usage_with_sys_argv(capfd, use_readline):
    with patch("sys.argv", new=["path.py", "--banner=test"] + use_readline):
        with patch("sys.stdin", new=io.StringIO("1+1\n")):
            with pytest.raises(SystemExit):
                apython.run_apython()
    out, err = capfd.readouterr()
    assert out == ""
    assert err == "test\n>>> 2\n>>> \n"


def test_apython_with_prompt_control(capfd):
    with patch("sys.stdin", new=io.StringIO("1+1\n")):
        with pytest.raises(SystemExit):
            apython.run_apython(
                ["--banner=test", "--prompt-control=▲", "--no-readline"]
            )
    out, err = capfd.readouterr()
    assert out == ""
    assert err == "test\n▲>>> ▲2\n▲>>> ▲\n"


def test_apython_with_prompt_control_and_ainput(capfd):
    input_string = "await ainput()\nhello\n"
    with patch("sys.stdin", new=io.StringIO(input_string)):
        with pytest.raises(SystemExit):
            apython.run_apython(
                ["--no-readline", "--banner=test", "--prompt-control=▲"]
            )
    out, err = capfd.readouterr()
    assert out == ""
    assert err == "test\n▲>>> ▲▲▲'hello'\n▲>>> ▲\n"


def test_apython_with_ainput(capfd, use_readline):
    input_string = "await ainput()\nhello\n"
    with patch("sys.stdin", new=io.StringIO(input_string)):
        with pytest.raises(SystemExit):
            apython.run_apython(["--banner=test"] + use_readline)
    out, err = capfd.readouterr()
    assert out == ""
    assert err == "test\n>>> 'hello'\n>>> \n"


def test_apython_with_stdout_logs(capfd, use_readline):
    with patch(
        "sys.stdin", new=io.StringIO('import sys; sys.stdout.write("logging") or 7\n')
    ):
        with pytest.raises(SystemExit):
            apython.run_apython(["--banner=test"] + use_readline)
    out, err = capfd.readouterr()
    assert out == "logging"
    assert err == "test\n>>> 7\n>>> \n"


def test_apython_server(capfd):
    def run_forever(self, orig=InteractiveEventLoop.run_forever):
        if self.console_server is not None:
            self.call_later(0, self.stop)
        return orig(self)

    with patch("aioconsole.InteractiveEventLoop.run_forever", run_forever):
        with pytest.raises(SystemExit):
            apython.run_apython(["--serve=:0"])
    out, err = capfd.readouterr()
    assert out.startswith("The console is being served on")
    assert err == ""


@pytest.mark.skipif(
    compat.platform == "win32", reason="apython does not run in a subprocess on windows"
)
def test_apython_non_existing_file(capfd):
    with pytest.raises(SystemExit):
        apython.run_apython(["idontexist.py"])
    out, err = capfd.readouterr()
    assert out == ""
    assert "No such file or directory" in err
    assert "idontexist.py" in err


@pytest.mark.skipif(
    compat.platform == "win32", reason="apython does not run in a subprocess on windows"
)
def test_apython_non_existing_module(capfd):
    with pytest.raises(SystemExit):
        apython.run_apython(["-m", "idontexist"])
    out, err = capfd.readouterr()
    assert out == ""
    assert "No module named idontexist" in err


def test_apython_pythonstartup(capfd, use_readline, monkeypatch, tmpdir):
    python_startup = tmpdir / "python_startup.py"
    monkeypatch.setenv("PYTHONSTARTUP", str(python_startup))
    python_startup.write(startupfile)

    test_vectors = (
        ("print(foo)\n", "", ">>> 1\n"),
        ("print(hehe())\n", "", ">>> 42\n"),
        ("print(r)\n", "", ">>> 1.0\n"),
        ("pprint({1:2})\n", "{1: 2}\n", ">>> >>> \n"),
    )
    inputstr = "".join([tv[0] for tv in test_vectors])
    outstr = "".join([tv[1] for tv in test_vectors])
    errstr = "test\n" + "".join([tv[2] for tv in test_vectors])

    with patch("sys.stdin", new=io.StringIO(inputstr)):
        with pytest.raises(SystemExit):
            apython.run_apython(["--banner=test"] + use_readline)
    out, err = capfd.readouterr()
    assert out == outstr
    assert err == errstr



"""Provide an interactive event loop class."""

import asyncio
import functools

from . import server
from . import console


class InteractiveEventLoop(asyncio.SelectorEventLoop):
    """Event loop running a python console."""

    console_class = console.AsynchronousConsole

    def __init__(
        self,
        *,
        selector=None,
        locals=None,
        banner=None,
        serve=None,
        prompt_control=None,
    ):
        self.console = None
        self.console_task = None
        self.console_server = None
        super().__init__(selector=selector)
        # Factory
        self.factory = lambda streams: self.console_class(
            streams, locals=locals, prompt_control=prompt_control, loop=self
        )
        # Local console
        if serve is None:
            self.console = self.factory(None)
            coro = self.console.interact(banner, stop=True, handle_sigint=True)
            self.console_task = asyncio.ensure_future(coro, loop=self)
        # Serving console
        else:
            host, port = serve
            coro = server.start_interactive_server(
                self.factory, host=host, port=port, banner=banner, loop=self
            )
            self.console_server = self.run_until_complete(coro)
            server.print_server(self.console_server)

    def close(self):
        if self.console_task and not self.is_running():
            asyncio.Future.cancel(self.console_task)
        super().close()

    def __del__(self):
        if self.console_task and self.console_task.done():
            self.console_task.exception()
        try:
            super().__del__()
        except AttributeError:
            pass


class InteractiveEventLoopPolicy(asyncio.DefaultEventLoopPolicy):
    """Policy to use the interactive event loop by default."""

    def __init__(self, *, locals=None, banner=None, serve=None, prompt_control=None):
        self._loop_factory = functools.partial(
            InteractiveEventLoop,
            locals=locals,
            banner=banner,
            serve=serve,
            prompt_control=prompt_control,
        )
        super().__init__()


def set_interactive_policy(
    *, locals=None, banner=None, serve=None, prompt_control=None
):
    """Use an interactive event loop by default."""
    policy = InteractiveEventLoopPolicy(
        locals=locals, banner=banner, serve=serve, prompt_control=prompt_control
    )
    asyncio.set_event_loop_policy(policy)


def run_console(*, locals=None, banner=None, serve=None, prompt_control=None):
    """Run the interactive event loop."""
    loop = InteractiveEventLoop(
        locals=locals, banner=banner, serve=serve, prompt_control=prompt_control
    )
    asyncio.set_event_loop(loop)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass



import io
import asyncio

import pytest
from aioconsole import aexec, aeval
from aioconsole.execute import compile_for_aexec


# Helper


async def coro(x):
    await asyncio.sleep(0)
    return x


# Test data

testdata = {
    "simple": ({}, "1+1", 2, {}),
    "affect": ({}, "a=1", None, {"a": 1}),
    "modify": ({"b": 1}, "b=2", None, {"b": 2}),
    "multiple": ({}, "c=3;d=4;5", 5, {"c": 3, "d": 4}),
    "async": ({"coro": coro}, "await coro(6)", 6, {"coro": coro}),
    "multiline string": ({}, '"""\n"""', "\n", {}),
    "splitted string": ({}, '"a\\\nb"', "ab", {}),
}


# Parametrized test for aexec


@pytest.mark.parametrize(
    "local, code, expected_result, expected_local",
    list(testdata.values()),
    ids=list(testdata.keys()),
)
@pytest.mark.asyncio
async def test_aexec_exec_mode(local, code, expected_result, expected_local):
    stream = io.StringIO()
    await aexec(code, local=local, stream=stream)
    assert stream.getvalue() == ""
    assert local == expected_local


@pytest.mark.parametrize(
    "local, code, expected_result, expected_local",
    list(testdata.values()),
    ids=list(testdata.keys()),
)
@pytest.mark.asyncio
async def test_aexec_single_mode(local, code, expected_result, expected_local):
    stream = io.StringIO()
    await aexec(compile_for_aexec(code, "test", "single"), local=local, stream=stream)
    if expected_result is None:
        assert stream.getvalue() == ""
    else:
        assert stream.getvalue().strip() == repr(expected_result)
    assert local.pop("_") == expected_result
    assert local == expected_local


def test_invalid_compile_modes():
    with pytest.raises(ValueError):
        compile_for_aexec("1\n", "test", "eval")
    with pytest.raises(ValueError):
        compile_for_aexec("1\n", "test", "unknown")


@pytest.mark.asyncio
async def test_incomplete_code():
    with pytest.raises(SyntaxError):
        await aexec("(")


@pytest.mark.asyncio
async def test_missing_newline():
    # Missing newline in "exec" mode is OK
    dct = {}
    await aexec("def f():\n  return 1", dct)
    assert dct["f"]() == 1
    await aexec("def g(): return 2", dct)
    assert dct["g"]() == 2

    # Missing newline in "single" raises a SyntaxError
    with pytest.raises(SyntaxError):
        await aexec(
            compile_for_aexec(
                "def f():\n    return 1", "test", "single", dont_imply_dedent=True
            )
        )


# Test return and yield handling


@pytest.mark.asyncio
async def test_return_handling():
    with pytest.raises(SyntaxError) as context:
        await aexec("return None, {}")

    assert context.value.msg == "'return' outside function"


@pytest.mark.asyncio
async def test_yield_handling():
    with pytest.raises(SyntaxError) as context:
        await aexec("for i in range(5): yield i")

    assert context.value.msg == "'yield' outside function"


@pytest.mark.asyncio
async def test_yield_from_handling():
    with pytest.raises(SyntaxError) as context:
        await aexec("yield from range(5)")

    assert context.value.msg == "'yield' outside function"


@pytest.mark.asyncio
async def test_correct():
    await aexec("def x(): return")
    await aexec("async def x(): return")
    await aexec("def x(): yield")
    await aexec("async def x(): yield")


def echo(x):
    """Sync function for aeval parameterized test."""
    return x


async def aecho(x):
    """Async function for aeval parameterized test."""
    return echo(x)


# Parametrized test with a variety of expressions
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expression, local",
    [
        # Valid Simple Expressions
        ("1 + 2", None),
        ("sum([i * i for i in range(10)])", None),
        # Invalid Expressions
        ("def foo(): return 42", None),
        ("x = 1", None),
        ("x = 1\nx + 1", None),
        ("for i in range(10): pass", None),
        ("if True: pass", None),
        ("while True: break", None),
        ("try: pass\nexcept: pass", None),
        # Expressions Involving Undefined Variables
        ("undefined_variable", None),
        ("undefined_function()", None),
        # Expressions with Deliberate Errors
        ("1/0", None),
        ("open('nonexistent_file.txt')", None),
        # Lambda and Anonymous Functions
        ("(lambda x: x * 2)(5)", None),
        # Expressions with Built-in Functions
        ("len('test')", None),
        ("min([3, 1, 4, 1, 5, 9])", None),
        ("max([x * x for x in range(10)])", None),
        # Boolean and Conditional Expressions
        ("True and False", None),
        ("not True", None),  # Boolean negation
        ("5 if True else 10", None),
        # String Manipulation
        ("'hello' + ' ' + 'world'", None),
        ("f'hello {42}'", None),
        # Complex List Comprehensions
        ("[x for x in range(5)]", None),
        ("[x * x for x in range(10) if x % 2 == 0]", None),
        # Expressions with Syntax Errors
        ("return 42", None),
        ("yield 5", None),
        # Test with await
        ("await aecho(5)", {"aecho": aecho, "echo": echo}),
        # Test invalid local
        ("...", []),
        ("...", "string_instead_of_dict"),
        ("...", 42),
        ("...", set()),
        ("...", ...),
        ("...", 1.5),
        ("...", object()),
        ("...", asyncio),
        ("...", lambda: ...),
        ("...", {"__result__": 99}),
        # Invalid expressions
        ("", None),
        (None, None),
        (0, None),
        ({}, None),
        (object(), None),
        (asyncio, None),
        (..., None),
        (lambda: ..., None),
        # Conflicting name in local
        ("x", {"x": 1, "__aeval_result__": 99}),
    ],
)
async def test_aeval(expression, local):

    async def capture(func, *args, **kwargs):
        try:
            if asyncio.iscoroutinefunction(func):
                result = await func(*args, **kwargs)
            else:
                result = func(*args, **kwargs)
        except Exception as exc:
            return (type(exc), None)
        else:
            return (None, result)

    # Remove the await keyword from the expression for the synchronous evaluation
    sync_expression = (
        expression.replace("await a", "") if isinstance(expression, str) else expression
    )

    # Capture and compare the results of the synchronous and asynchronous evaluations
    sync_capture = await capture(eval, sync_expression, local)
    async_capture = await capture(aeval, expression, local)
    assert sync_capture == async_capture


# Test calling an async function without awaiting it
@pytest.mark.asyncio
async def test_aeval_async_func_without_await():
    expression = "asyncio.sleep(0)"
    local = {"asyncio": asyncio}
    result = await aeval(expression, local)
    assert asyncio.iscoroutine(result)
    await result


@pytest.mark.asyncio
async def test_aeval_valid_await_syntax():
    expression = "await aecho(10)"
    local = {"aecho": aecho}
    result = await aeval(expression, local)
    assert result == 10


@pytest.mark.asyncio
async def test_aeval_coro_in_local():
    expression = "await coro"
    local = {"coro": aecho(10)}
    result = await aeval(expression, local)
    assert result == 10



"""Provide asynchronous equivalent to input, exec and interact.

It also includes an interactive event loop, and a command line interface.
"""

from .execute import aexec, aeval
from .console import AsynchronousConsole, interact
from .stream import ainput, aprint, get_standard_streams
from .events import InteractiveEventLoop, InteractiveEventLoopPolicy
from .events import set_interactive_policy, run_console
from .command import AsynchronousCli
from .server import start_interactive_server
from .apython import run_apython


__version__ = "0.8.0"

__all__ = [
    "aexec",
    "aeval",
    "ainput",
    "aprint",
    "AsynchronousConsole",
    "interact",
    "InteractiveEventLoop",
    "InteractiveEventLoopPolicy",
    "set_interactive_policy",
    "run_console",
    "AsynchronousCli",
    "start_interactive_server",
    "get_standard_streams",
    "run_apython",
]



"""Provide an asynchronous equivalent to the python console."""

import sys
import code
import pydoc
import codeop
import signal
import asyncio
import inspect
import functools
import traceback

from . import stream
from . import execute

EXTRA_MESSAGE = """\
---
This console is running in an asyncio event loop.
It allows you to wait for coroutines using the 'await' syntax.
Try: await asyncio.sleep(1, result=3)
---"""

# cx_Freeze does not include the help function
try:
    help_function = help
except NameError:
    help_function = None


class AsynchronousCompiler(codeop.CommandCompiler):
    def __init__(self):
        self.compiler = functools.partial(
            execute.compile_for_aexec, dont_imply_dedent=True
        )


class AsynchronousConsole(code.InteractiveConsole):
    def __init__(
        self,
        streams=None,
        locals=None,
        filename="<console>",
        prompt_control=None,
        *,
        loop=None,
    ):
        super().__init__(locals, filename)
        # Process arguments
        if loop is None:
            loop = asyncio.get_event_loop()
        if streams is None:
            streams = stream.get_standard_streams(use_stderr=True, loop=loop)
        # Attributes
        self.streams = streams
        self.loop = loop
        self.reader = None
        self.writer = None
        self.prompt_control = prompt_control
        self.compile = AsynchronousCompiler()
        # Populate locals
        self.locals["asyncio"] = asyncio
        self.locals["loop"] = self.loop
        self.locals["print"] = self.print
        self.locals["help"] = self.help
        self.locals["ainput"] = self.ainput
        # Internals
        self._sigint_received = False

    @functools.wraps(print)
    def print(self, *args, **kwargs):
        kwargs.setdefault("file", self)
        print(*args, **kwargs)

    @functools.wraps(help_function)
    def help(self, obj):
        self.print(pydoc.render_doc(obj))

    @functools.wraps(stream.ainput)
    async def ainput(self, prompt="", *, streams=None, use_stderr=False, loop=None):
        # Get the console streams by default
        if streams is None and use_stderr is False:
            streams = self.reader, self.writer
        # Wrap the prompt with prompt control characters
        if self.prompt_control and self.prompt_control not in prompt:
            prompt = self.prompt_control + prompt + self.prompt_control
        # Run ainput
        return await stream.ainput(
            prompt, streams=streams, use_stderr=use_stderr, loop=loop
        )

    def get_default_banner(self):
        cprt = (
            'Type "help", "copyright", "credits" ' 'or "license" for more information.'
        )
        return f"Python {sys.version} on {sys.platform}\n{cprt}\n{EXTRA_MESSAGE}"

    async def runsource(self, source, filename="<ainput>", symbol="single"):
        try:
            code = self.compile(source, filename, symbol)
        except (OverflowError, SyntaxError, ValueError):
            self.showsyntaxerror(filename)
            return False

        if code is None:
            return True

        await self.runcode(code)
        return False

    async def runcode(self, code):
        try:
            await execute.aexec(
                code, local=self.locals, stream=self, filename=self.filename
            )
        except SystemExit:
            raise
        except BaseException:
            self.showtraceback()
        await self.flush()

    def resetbuffer(self):
        self.buffer = []

    def handle_sigint(self, task):
        self._sigint_received = True
        task.cancel()
        if task._fut_waiter._loop is not self.loop:
            task._wakeup(task._fut_waiter)

    def add_sigint_handler(self):
        task = asyncio.current_task(loop=self.loop)
        try:
            self.loop.add_signal_handler(signal.SIGINT, self.handle_sigint, task)
        except NotImplementedError:

            def callback(*args):
                self.loop.call_soon_threadsafe(self.handle_sigint, task)

            signal.signal(signal.SIGINT, callback)

    def remove_sigint_handler(self):
        try:
            self.loop.remove_signal_handler(signal.SIGINT)
        except NotImplementedError:
            signal.signal(signal.SIGINT, signal.default_int_handler)

    async def interact(self, banner=None, stop=True, handle_sigint=True):
        # Get the streams
        try:
            if inspect.isawaitable(self.streams):
                self.reader, self.writer = await self.streams
            elif self.streams is not None:
                self.reader, self.writer = self.streams
        finally:
            self.streams = None
        # Interact
        try:
            if handle_sigint:
                self.add_sigint_handler()
            await self._interact(banner)
            if stop:
                raise SystemExit
        # Exit
        except SystemExit:
            if stop:
                raise
        # Clean-up
        finally:
            if handle_sigint:
                self.remove_sigint_handler()

    async def _interact(self, banner=None):
        # Get ps1 and ps2
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        # Print banner
        if banner is None:
            banner = self.get_default_banner()
        self.write(f"{banner}\n")
        # Run loop
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = await self.raw_input(prompt)
                except EOFError:
                    try:
                        self.write("\n")
                        await self.flush()
                    except OSError:
                        pass
                    break
                else:
                    more = await self.push(line)
            except asyncio.CancelledError:
                # Not our cancellation
                if not self._sigint_received:
                    raise
                # Manage cancellation
                self._sigint_received = False
                self.write("\nKeyboardInterrupt\n")
                await self.flush()
                self.resetbuffer()
                more = 0

    async def push(self, line):
        self.buffer.append(line)
        source = "\n".join(self.buffer)
        more = await self.runsource(source, self.filename)
        if not more:
            self.resetbuffer()
        return more

    async def raw_input(self, prompt=""):
        return await self.ainput(prompt)

    def write(self, data):
        return self.writer.write(data.encode())

    async def flush(self):
        try:
            await self.writer.drain()
        except ConnectionResetError:
            pass

    # Re-implement showtraceback and showsyntaxerror
    # to ignore sys.excepthook (set by ubuntu apport for instance)

    def showtraceback(self):
        sys.last_type, sys.last_value, last_tb = ei = sys.exc_info()
        sys.last_traceback = last_tb
        try:
            lines = traceback.format_exception(ei[0], ei[1], last_tb.tb_next)
            self.write("".join(lines))
        finally:
            last_tb = ei = None

    def showsyntaxerror(self, filename=None):
        type, value, tb = sys.exc_info()
        sys.last_type = type
        sys.last_value = value
        sys.last_traceback = tb
        if filename and type is SyntaxError:
            # Work hard to stuff the correct filename in the exception
            try:
                msg, (dummy_filename, lineno, offset, line) = value.args
            except ValueError:
                # Not the format we expect; leave it alone
                pass
            else:
                # Stuff in the right filename
                value = SyntaxError(msg, (filename, lineno, offset, line))
                sys.last_value = value
        lines = traceback.format_exception_only(type, value)
        self.write("".join(lines))


async def interact(
    banner=None,
    streams=None,
    locals=None,
    prompt_control=None,
    stop=True,
    handle_sigint=True,
    *,
    loop=None,
):
    console = AsynchronousConsole(
        streams, locals=locals, prompt_control=prompt_control, loop=loop
    )
    await console.interact(banner, stop, handle_sigint)
